import inspect
from collections.abc import AsyncGenerator, AsyncIterable, Callable, Generator, Iterable
from typing import Any, overload

from peritype import FWrap, TWrap, wrap_func, wrap_type
from peritype.collections import TypeBag, TypeMap

from soupape.resolvers import ServiceDefaultResolver
from soupape.types import InjectionScope, ServiceResolver, TypeResolverMetadata


class ServiceCollection:
    def __init__(self) -> None:
        self._registered_services = TypeBag()
        self._resolvers = TypeMap[Any, TypeResolverMetadata[..., Any]]()

    def _add_resolver(self, metadata: TypeResolverMetadata[..., Any]) -> None:
        if metadata.interface in self._registered_services:
            raise ValueError(f"Service resolver for type {metadata.interface} is already registered.")
        self._registered_services.add(metadata.interface)
        self._resolvers.add(metadata.interface, metadata)

    def _unpack_resolver_function_return(self, func: FWrap[..., Any]) -> TWrap[Any]:
        original = func.func
        hint = func.get_return_hint()
        if inspect.isasyncgenfunction(original):
            if hint.match(AsyncGenerator[Any, Any] | AsyncIterable[Any]):
                return hint.generic_params[0]
            else:
                raise TypeError("Async generator resolver functions must have return type hint of AsyncGenerator[T].")
        elif inspect.isgeneratorfunction(original):
            if hint.match(Iterable[Any] | Generator[Any, Any, Any]):
                return hint.generic_params[0]
            else:
                raise TypeError("Generator resolver functions must have return type hint of Iterable[T].")
        return hint

    def _unpack_registration_args(
        self,
        scope: InjectionScope,
        args: tuple[Any, ...],
    ) -> TypeResolverMetadata[..., Any]:
        implementation: type[Any] | None = None
        interface: type[Any] | None = None
        match args:
            case (type() as i1,):
                implementation = i1
                interface = i1
                resolver = None
            case (type() as i1, type() as i2):
                interface = i1
                implementation = i2
                resolver = None
            case (Callable() as r1,):
                interface = None
                implementation = None
                resolver = r1
            case _:
                raise TypeError()

        if interface is None or implementation is None:
            assert resolver is not None
            fwrap = wrap_func(resolver)
            wrap_interface = self._unpack_resolver_function_return(fwrap)
            wrap_implementation = wrap_interface
            signature = fwrap.signature
        else:
            assert resolver is None
            wrap_interface = wrap_type(interface)
            wrap_implementation = wrap_type(implementation)
            resolver = ServiceDefaultResolver(self, wrap_interface, wrap_implementation)
            signature = wrap_implementation.signature
            fwrap = wrap_interface.init

        return TypeResolverMetadata(
            scope=scope,
            interface=wrap_interface,
            implementation=wrap_implementation,
            resolver=resolver,
            signature=signature,
            fwrap=fwrap,
        )

    def is_registered[T](self, interface: type[T] | TWrap[T]) -> bool:
        if not isinstance(interface, TWrap):
            interface = wrap_type(interface)
        return interface in self._registered_services

    def get_metadata[T](self, interface: type[T] | TWrap[T]) -> TypeResolverMetadata[..., T]:
        if not isinstance(interface, TWrap):
            interface = wrap_type(interface)
        return self._resolvers[interface]

    @overload
    def add_singleton[IntrT, ImplT](self, interface: type[IntrT], implementation: type[ImplT], /) -> None: ...
    @overload
    def add_singleton[ImplT](self, implementation: type[ImplT], /) -> None: ...
    @overload
    def add_singleton[**P, IntrT](self, resolver: ServiceResolver[P, IntrT], /) -> None: ...
    def add_singleton(self, *args: Any) -> None:
        metadata = self._unpack_registration_args(InjectionScope.SINGLETON, args)
        self._add_resolver(metadata)

    @overload
    def add_scoped[IntrT, ImplT](self, interface: type[IntrT], implementation: type[ImplT], /) -> None: ...
    @overload
    def add_scoped[ImplT](self, implementation: type[ImplT], /) -> None: ...
    @overload
    def add_scoped[**P, IntrT](self, resolver: ServiceResolver[P, IntrT], /) -> None: ...
    def add_scoped(self, *args: Any) -> None:
        metadata = self._unpack_registration_args(InjectionScope.SCOPED, args)
        self._add_resolver(metadata)

    @overload
    def add_transient[IntrT, ImplT](self, interface: type[IntrT], implementation: type[ImplT], /) -> None: ...
    @overload
    def add_transient[ImplT](self, implementation: type[ImplT], /) -> None: ...
    @overload
    def add_transient[**P, IntrT](self, resolver: ServiceResolver[P, IntrT], /) -> None: ...
    def add_transient(self, *args: Any) -> None:
        metadata = self._unpack_registration_args(InjectionScope.TRANSIENT, args)
        self._add_resolver(metadata)

    def copy(self) -> "ServiceCollection":
        new_collection = ServiceCollection()
        new_collection._registered_services = self._registered_services.copy()
        new_collection._resolvers = self._resolvers.copy()
        return new_collection



import asyncio
import inspect
from collections.abc import AsyncGenerator, Callable, Coroutine
from typing import Any, Self, Unpack, cast, overload

from peritype import FWrap, TWrap, wrap_func, wrap_type

from soupape.collection import ServiceCollection
from soupape.injector import BaseInjector
from soupape.instances import InstancePoolStack
from soupape.types import InjectionScope, Injector, InjectorCallArgs, ResolverCallArgs, ResolverMetadata


class AsyncInjector(BaseInjector, Injector):
    def __init__(self, services: ServiceCollection, instance_pool: InstancePoolStack | None = None) -> None:
        super().__init__(services, instance_pool)
        self._async_generators_to_close: list[AsyncGenerator[Any]] = []
        self._set_injector_in_services()

    @property
    def is_async(self) -> bool:
        return True

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: Any,
    ) -> None:
        for gen in self._generators_to_close:
            try:
                next(gen)
            except StopIteration:
                pass
        for agen in self._async_generators_to_close:
            try:
                await anext(agen)
            except StopAsyncIteration:
                pass

    def _set_injector_in_services(self) -> None:
        injector_w = wrap_type(AsyncInjector)
        if not self._services.is_registered(injector_w):

            def resolver() -> "AsyncInjector":
                return self

            self._services.add_singleton(resolver)
        self._instance_pool.set_instance(injector_w, self)

    async def _resolve_service[T](
        self,
        resolver_args: ResolverCallArgs[..., T],
        *,
        injector_args: InjectorCallArgs | None = None,
    ) -> T:
        if resolver_args.interface is not None and self._has_instance(resolver_args.interface):
            return self._get_instance(resolver_args.interface)

        resolved_args: list[Any] = []
        if injector_args is not None and "positional_args" in injector_args:
            for arg in injector_args["positional_args"]:
                resolved_args.append(arg)
        for arg in resolver_args.args:
            resolved_arg = await self._resolve_service(arg)
            resolved_args.append(resolved_arg)

        resolved_kwargs: dict[str, Any] = {}
        for kwarg_name, kwarg in resolver_args.kwargs.items():
            resolved_kwarg = await self._resolve_service(kwarg)
            resolved_kwargs[kwarg_name] = resolved_kwarg

        resolver = self._get_resolver_from_call_args(resolver_args)
        resolved = resolver(*resolved_args, **resolved_kwargs)

        if inspect.isgenerator(resolved):
            self._generators_to_close.append(resolved)
            resolved = next(resolved)
        elif inspect.isasyncgen(resolved):
            self._async_generators_to_close.append(resolved)
            resolved = await anext(resolved)
        if inspect.iscoroutine(resolved):
            resolved = await resolved

        if resolver_args.interface is not None:
            self._set_instance(resolver_args.scope, resolver_args.interface, resolved)

        return resolved  # type: ignore

    async def require[T](self, interface: type[T] | TWrap[T]) -> T:
        if not isinstance(interface, TWrap):
            twrap = wrap_type(interface)
        else:
            twrap = interface
        return await self._require(twrap)

    async def _require[T](self, interface: TWrap[T]) -> T:
        metadata = self._get_service_metadata(interface)
        resolver_args = self._build_resolve_tree(metadata)
        resolved = await self._resolve_service(resolver_args)
        return resolved

    @overload
    async def call[**P, T](
        self,
        callable: FWrap[P, Coroutine[Any, Any, T]],
        **kwargs: Unpack[InjectorCallArgs],
    ) -> T: ...
    @overload
    async def call[**P, T](
        self,
        callable: FWrap[P, T],
        **kwargs: Unpack[InjectorCallArgs],
    ) -> T: ...
    @overload
    async def call[**P, T](
        self,
        callable: Callable[P, Coroutine[Any, Any, T]],
        **kwargs: Unpack[InjectorCallArgs],
    ) -> T: ...
    @overload
    async def call[**P, T](
        self,
        callable: Callable[P, T],
        **kwargs: Unpack[InjectorCallArgs],
    ) -> T: ...
    async def call(
        self,
        callable: Callable[..., Any] | FWrap[..., Any],
        **kwargs: Unpack[InjectorCallArgs],
    ) -> Any:
        if not isinstance(callable, FWrap):
            fwrap = wrap_func(callable)
        else:
            fwrap = cast(FWrap[..., Any], callable)

        async def resolver(*args: Any, **kwds: Any) -> Any:
            resolved = fwrap.func(*args, **kwds)
            if asyncio.iscoroutine(resolved):
                return await resolved
            return resolved

        resolver_args = self._build_resolve_tree(
            ResolverMetadata(
                scope=InjectionScope.IMMEDIATE,
                signature=fwrap.signature,
                fwrap=fwrap,
                resolver=resolver,
            ),
            injector_args=kwargs,
        )
        return await self._resolve_service(resolver_args, injector_args=kwargs)

    def get_scoped_injector(self) -> "AsyncInjector":
        return AsyncInjector(self._services, self._instance_pool.stack())



from collections.abc import AsyncIterable, Awaitable
from typing import Any


class SoupapeError(Exception):
    def __init__(self, code: str, message: str) -> None:
        super().__init__(f"[{code}] {message}")
        self.code = code
        self.message = message


class ServiceNotFoundError(SoupapeError):
    def __init__(self, interface: str) -> None:
        super().__init__(
            "soupape.service.not_found",
            f"Service for interface '{interface}' not found.",
        )


class MissingTypeHintError(SoupapeError):
    def __init__(self, parameter: str, fwrap: str) -> None:
        super().__init__(
            "soupape.type_hint.missing",
            f"Missing type hint for parameter '{parameter}' of '{fwrap}'.",
        )


class ScopedServiceNotAvailableError(SoupapeError):
    def __init__(self, interface: str) -> None:
        super().__init__(
            "soupape.scoped_service.not_available",
            f"Scoped service for interface '{interface}' is not available in the root scope.",
        )


class AsyncInSyncInjectorError(SoupapeError):
    def __init__(self, coro: Awaitable[Any] | AsyncIterable[Any]) -> None:
        super().__init__(
            "soupape.injector.async_in_sync",
            "Cannot call asynchronous resolver in synchronous injector.",
        )
        self._coro = coro

    async def close(self) -> None:
        if isinstance(self._coro, Awaitable):
            await self._coro



from collections.abc import Awaitable, Callable
from typing import Any, overload

from soupape.metadata import meta


class PostInitMetadata:
    KEY = "post_init"


@overload
def post_init[**P](func: Callable[P, Awaitable[None]]) -> Callable[P, Awaitable[None]]: ...
@overload
def post_init[**P](func: Callable[P, None]) -> Callable[P, None]: ...
def post_init(func: Callable[..., Any]) -> Callable[..., Any]:
    meta.set(func, PostInitMetadata.KEY, PostInitMetadata())
    return func



from collections.abc import Callable, Iterable
from typing import TYPE_CHECKING, Any

from peritype import TWrap

from soupape.metadata import meta
from soupape.post_init import PostInitMetadata
from soupape.resolvers import AsyncServiceDefaultResolver, SyncServiceDefaultResolver
from soupape.types import Injector, ServiceResolver, ServiceResolverFactory

if TYPE_CHECKING:
    from soupape.collection import ServiceCollection


class ServiceDefaultResolver[T](ServiceResolverFactory):
    def __init__(
        self,
        services: "ServiceCollection",
        interface: TWrap[T],
        implementation: TWrap[Any],
    ) -> None:
        self._services = services
        self._interface = interface
        self._implementation = implementation

    @property
    def interface(self) -> TWrap[T]:
        return self._interface

    @property
    def implementation(self) -> TWrap[Any]:
        return self._implementation

    def get_post_inits(self, twrap: TWrap[Any]) -> Iterable[Callable[..., Any]]:
        inner_type: type[Any] = twrap.inner_type
        for attr in vars(inner_type).values():
            if callable(attr) and meta.has(attr, PostInitMetadata.KEY):
                yield attr

    def with_injector(self, injector: Injector) -> ServiceResolver[..., T]:
        if injector.is_async:
            return AsyncServiceDefaultResolver(self, injector)
        else:
            return SyncServiceDefaultResolver(self, injector)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        raise NotImplementedError()



import inspect
from collections.abc import AsyncGenerator, AsyncIterable, Awaitable, Callable, Coroutine, Generator, Iterable
from dataclasses import dataclass
from enum import Enum, auto, unique
from types import TracebackType
from typing import Any, NotRequired, Protocol, TypedDict, Unpack, runtime_checkable

from peritype import FWrap, TWrap

type ServiceResolver[**P, T] = (
    Callable[P, T]
    | Callable[P, Generator[T]]
    | Callable[P, Iterable[T]]
    | Callable[P, AsyncGenerator[T]]
    | Callable[P, AsyncIterable[T]]
    | Callable[P, Coroutine[Any, Any, T]]
    | Callable[P, Awaitable[T]]
)


class InjectorCallArgs(TypedDict):
    positional_args: NotRequired[list[Any]]


class Injector(Protocol):
    @property
    def is_async(self) -> bool: ...

    def require[T](self, interface: type[T] | TWrap[T]) -> T | Awaitable[T]: ...

    def call[T](
        self,
        callable: Callable[..., T] | FWrap[..., T],
        **kwargs: Unpack[InjectorCallArgs],
    ) -> T | Awaitable[T]: ...

    def get_scoped_injector(self) -> "Injector": ...


@runtime_checkable
class ServiceResolverFactory(Protocol):
    def with_injector(self, injector: Injector) -> "ServiceResolver[..., Any]": ...


@runtime_checkable
class SyncContextManager(Protocol):
    def __enter__(self) -> "SyncContextManager": ...

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None: ...


@runtime_checkable
class AsyncContextManager(Protocol):
    async def __aenter__(self) -> "AsyncContextManager": ...

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None: ...


@unique
class InjectionScope(Enum):
    SINGLETON = auto()
    SCOPED = auto()
    TRANSIENT = auto()
    IMMEDIATE = auto()


@dataclass
class ResolverMetadata[**P, T]:
    scope: InjectionScope
    signature: inspect.Signature
    fwrap: FWrap[P, T]
    resolver: ServiceResolver[P, T]


@dataclass
class TypeResolverMetadata[**P, T](ResolverMetadata[P, T]):
    interface: TWrap[T]
    implementation: TWrap[Any]


@dataclass
class ResolverCallArgs[**P, T]:
    scope: InjectionScope
    args: "list[ResolverCallArgs[..., Any]]"
    kwargs: "dict[str, ResolverCallArgs[..., Any]]"
    resolver: ServiceResolver[P, T]
    interface: TWrap[T] | None



from soupape.resolvers.async_default import AsyncServiceDefaultResolver as AsyncServiceDefaultResolver
from soupape.resolvers.sync_default import SyncServiceDefaultResolver as SyncServiceDefaultResolver
from soupape.resolvers.default import ServiceDefaultResolver as ServiceDefaultResolver



from soupape.types import Injector as Injector
from soupape.collection import ServiceCollection as ServiceCollection
from soupape.post_init import post_init as post_init
from soupape.injector import AsyncInjector as AsyncInjector, SyncInjector as SyncInjector



from soupape.injector.base import BaseInjector as BaseInjector
from soupape.injector.async_injector import AsyncInjector as AsyncInjector
from soupape.injector.sync_injector import SyncInjector as SyncInjector



from collections.abc import Generator
from typing import Any, Self

from peritype import TWrap

from soupape import ServiceCollection
from soupape.errors import MissingTypeHintError, ScopedServiceNotAvailableError, ServiceNotFoundError
from soupape.instances import InstancePoolStack
from soupape.types import (
    InjectionScope,
    Injector,
    InjectorCallArgs,
    ResolverCallArgs,
    ResolverMetadata,
    ServiceResolver,
    ServiceResolverFactory,
    TypeResolverMetadata,
)


class BaseInjector(Injector):
    def __init__(self, services: ServiceCollection, instance_pool: InstancePoolStack | None = None) -> None:
        self._services = services.copy()
        self._instance_pool = instance_pool if instance_pool is not None else InstancePoolStack()
        self._generators_to_close: list[Generator[Any]] = []

    @property
    def is_root_injector(self) -> bool:
        return len(self._instance_pool) == 1

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: Any,
    ) -> None:
        for gen in self._generators_to_close:
            try:
                next(gen)
            except StopIteration:
                pass

    def _has_instance(self, twrap: TWrap[Any]) -> bool:
        return twrap in self._instance_pool

    def _set_instance(self, scope: InjectionScope, twrap: TWrap[Any], instance: Any) -> None:
        match scope:
            case InjectionScope.IMMEDIATE | InjectionScope.TRANSIENT:
                return
            case InjectionScope.SINGLETON:
                set_to_root = True
            case InjectionScope.SCOPED:
                if self.is_root_injector:
                    raise ScopedServiceNotAvailableError(str(twrap))
                set_to_root = False
        self._instance_pool.set_instance(twrap, instance, root=set_to_root)

    def _get_instance[InstanceT](self, twrap: TWrap[InstanceT]) -> InstanceT:
        return self._instance_pool.get_instance(twrap)

    def _get_service_metadata(self, interface: TWrap[Any]) -> TypeResolverMetadata[..., Any]:
        if not self._services.is_registered(interface):
            raise ServiceNotFoundError(str(interface))
        return self._services.get_metadata(interface)

    def _build_resolve_tree(
        self,
        metadata: ResolverMetadata[..., Any] | TypeResolverMetadata[..., Any],
        *,
        injector_args: InjectorCallArgs | None = None,
    ) -> ResolverCallArgs[..., Any]:
        args: list[ResolverCallArgs[..., Any]] = []
        kwargs: dict[str, ResolverCallArgs[..., Any]] = {}
        hints = metadata.fwrap.get_signature_hints()

        if injector_args is not None and "positional_args" in injector_args:
            skip = len(injector_args["positional_args"])
        else:
            skip = 0

        for param_name, param in metadata.signature.parameters.items():
            if skip > 0:
                skip -= 1
                continue
            if param_name not in hints:
                raise MissingTypeHintError(param_name, str(metadata.fwrap))
            hint = hints[param_name]
            hint_metadata = self._get_service_metadata(hint)
            resolver_args = self._build_resolve_tree(hint_metadata)
            if param.kind in (param.POSITIONAL_ONLY, param.POSITIONAL_OR_KEYWORD):
                args.append(resolver_args)
            elif param.kind == param.KEYWORD_ONLY:
                kwargs[param_name] = resolver_args

        return ResolverCallArgs(
            metadata.scope,
            args,
            kwargs,
            metadata.resolver,
            metadata.interface if isinstance(metadata, TypeResolverMetadata) else None,
        )

    def _get_resolver_from_call_args(self, call_args: ResolverCallArgs[..., Any]) -> ServiceResolver[..., Any]:
        if isinstance(call_args.resolver, ServiceResolverFactory):
            return call_args.resolver.with_injector(self)
        return call_args.resolver



import asyncio
from collections.abc import AsyncGenerator
from typing import TYPE_CHECKING, Any

from soupape.types import AsyncContextManager, Injector, SyncContextManager

if TYPE_CHECKING:
    from soupape.resolvers import ServiceDefaultResolver


class AsyncServiceDefaultResolver[T]:
    def __init__(self, resolver: "ServiceDefaultResolver[T]", injector: Injector) -> None:
        self._resolver = resolver
        self._injector = injector

    async def __call__(self, *args: Any, **kwargs: Any) -> AsyncGenerator[T]:
        instance = self._resolver.implementation.instantiate(*args, **kwargs)
        post_inits = self._resolver.get_post_inits(self._resolver.implementation)
        for post_init in post_inits:
            result = self._injector.call(post_init, positional_args=[instance])
            if asyncio.iscoroutine(result):
                await result
        if isinstance(instance, AsyncContextManager):
            async with instance:
                yield instance  # pyright: ignore[reportReturnType]
            return
        if isinstance(instance, SyncContextManager):
            with instance:
                yield instance  # pyright: ignore[reportReturnType]
            return
        yield instance
